/*
//
// BEGIN SONGBIRD GPL
//
// This file is part of the Songbird web player.
//
// Copyright(c) 2005-2008 POTI, Inc.
// http://songbirdnest.com
//
// This file may be licensed under the terms of of the
// GNU General Public License Version 2 (the "GPL").
//
// Software distributed under the License is distributed
// on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either
// express or implied. See the GPL for the specific language
// governing rights and limitations.
//
// You should have received a copy of the GPL along with this
// program. If not, go to http://www.gnu.org/licenses/gpl.html
// or write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
//
// END SONGBIRD GPL
//
*/


const Cc = Components.classes;
const Ci = Components.interfaces;
const Cr = Components.results;
const Ce = Components.Exception;
const Cu = Components.utils;


Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");
Components.utils.import("resource://app/jsmodules/sbProperties.jsm");


/*=========================================================================*/
/* Daap Client                                                             */
/*=========================================================================*/

/* constructor */
function sbDaapClient() {
  this.channel = null;
  this.uri = null;
  this._serverInfo = null;
}

/* implementation */
sbDaapClient.prototype = {
  /* nsISupports ---------------------------------------------------------*/
  QueryInterface: XPCOMUtils.generateQI([Ci.sbIDaapClient, 
                                          Ci.nsIClassInfo]),

  
  /* nsIClassInfo, so callers don't have to QI from JS -------------------*/
  classDescription: "Daap Client",
  classID:          Components.ID("{bc9a9f10-1928-4c87-8094-fe615921e9a3}"),
  contractID:       "@songbirdnest.com/daap/client;1",
  flags                   : Ci.nsIClassInfo.THREADSAFE,
  implementationLanguage  : Ci.nsIProgrammingLanguage.JAVASCRIPT,
  getHelperForLanguage    : function(aLanguage) { return null; },
  getInterfaces : function(count) {
    var interfaces = [Ci.sbIDaapClient, Ci.nsIClassInfo, Ci.nsISupports];
    count.value = interfaces.length;
    return interfaces;
  },


  /* private data ---------------------------------------------------------*/
  uri: null,
  channel: null,
  _serverInfo: null,


  /* sbIDaapClient --------------------------------------------------------*/

  /* readonly attribute long lastHttpChannel; */
  get lastHttpChannel () {
    if (this.channel) {
      return this.channel;
    }
    throw Cr.NS_ERROR_NOT_AVAILABLE;
  },

  /* readonly attribute nsIURI URI; */
  get URI () {
    if (this.uri) {
      return this.uri;
    }
    throw Cr.NS_ERROR_NOT_AVAILABLE;
  },
  
  /* void init (in nsIURI aURI); */
  init: function (aURI)
  {
    this.uri = aURI;
    this.channel = null;

    /* get the server info */
    this.getServerInfo();

    /* throws NS_ERROR_NOT_AVAILABLE on connection errors */

  },


  /* readonly attribute sbIDaapServerInfo serverInfo; */
  get serverInfo () {
    if (this._serverInfo) {
      return this._serverInfo;
    }
    return this.getServerInfo();
  },


  /* sbIDaapServerInfo getServerInfo (); */
  getServerInfo: function() 
  {
    var path = "/server-info";
    var handler = new SimpleHandler();

    var response = this.download(path , handler);
    if (!response){
      return null;
    }

    var result = new sbDaapServerInfo();
            
    if (response.mstt) result.infoStatus = response.mstt;
    if (response.apro) result.protocol = response.apro;
    if (response.msix) result.supportIndexing = response.msix;
    if (response.msex) result.supportExtensions = response.msex;
    if (response.msup) result.supportUpdate = response.msup;
    if (response.msal) result.supportAutoLogout = response.msal;
    if (response.mstm) result.timeoutInterval = response.mstm;
    if (response.mslr) result.loginRequired = response.mslr;
    if (response.msqy) result.supportQueries = response.msqy;
    if (response.minm) result.serverName = response.minm;
    if (response.msrs) result.supportResolve = response.msrs;
    if (response.msbr) result.supportBrowsing = response.msbr;
    if (response.mspi) result.supportPersistentID = response.mspi;
    if (response.mpro) result.dmapVersion = response.mpro;

    /* cache for the next time */
    this._serverInfo = result;

    return result;
  },

  /* nsIArray getContentCodes(); */
  getContentCodes: function ()
  {
    var path = "/content-codes";
    var handler = new ContainerHandler();
    handler.tag = "mdcl"; //search this tag

    var response = this.download(path , handler);
    if (!response){
      return null;
    }
    
    /* only care for the container records */
    var result = Cc["@mozilla.org/array;1"]
                  .createInstance(Ci.nsIMutableArray);

    for(var i in response){
      var record = new sbDaapContentCode();

      record.code = response[i].mcnm;
      record.name = response[i].mcna;
      record.type = response[i].mcty;

      result.appendElement(record, false);
    }

    return result;
  },

  /* long login ([optional] in boolean noInitSession); */
  login: function (aPassword, noInitSession) 
  {
    if (!this.uri) {
      throw Cr.NS_ERROR_NOT_INITIALIZED;
    }

    if (!noInitSession) {
      /* delegate the login to the session manager */
      var sessionManager = Cc["@songbirdnest.com/daap/session-manager;1"]
                            .getService(Ci.sbIDaapSessionManager);

      return sessionManager.initSession(this, aPassword);
    }

    /* do the request */
    var path = "/login";
    var handler = new SimpleHandler();

    var response = this.download(path , handler);
    if (!response){
      return null;
    }
    if (this.channel && this.channel.responseStatus != 200){
      return null;
    }

    return response.mlid;
  },

  /* void logout ([optional] in boolean noCloseSession); */
  logout: function(noCloseSession) 
  {
    if (!this.uri) {
      throw Cr.NS_ERROR_NOT_INITIALIZED;
    }

    if (!noCloseSession) {
      /* delegate the logout to the session manager */
      var sessionManager = Cc["@songbirdnest.com/daap/session-manager;1"]
                            .getService(Ci.sbIDaapSessionManager);

      return sessionManager.closeSession(this);
    }

    /* do the request */
    var path = "/logout";
    this.download(path , null);
  },

  /* long getServerRevision(); */
  getServerRevision: function ()
  {
    var path = "/update";
    var handler = new SimpleHandler();

    var response = this.download(path , handler);
    if (!response){
      return null;
    }

    var result = response.musr;
    return result;
  },

  /* nsIArray getDatabases (); */
  getDatabases: function() 
  {
    var path = "/databases";
    var handler = new ContainerHandler();

    var response = this.download(path , handler);
    if (!response){
      return null;
    }
    
    /* only care for the container records */
    var result = Cc["@mozilla.org/array;1"]
                  .createInstance(Ci.nsIMutableArray);

    for(var i in response){
      var record = new sbDaapDatabase();
      record.id = response[i].miid;
      record.persistentId = response[i].mper;
      record.name = response[i].minm;
      record.itemCount = response[i].mimc;
      record.containerCount = response[i].mctc;

      result.appendElement(record, false);
    }

    return result;
  },

  /* nsIArray getPlaylists (in long aDatabaseId); */
  getPlaylists: function(aDatabaseId) 
  {
    var path = "/databases/"+ aDatabaseId + "/containers";
    var handler = new ContainerHandler();

    var response = this.download(path, handler);
    if (!response){
      return null;
    }
    
    /* only care for the container records */
    var result = Cc["@mozilla.org/array;1"]
                  .createInstance(Ci.nsIMutableArray);

    for(var i in response){
      var record = new sbDaapPlaylist();
      record.id = response[i].miid;
      record.persistentId = response[i].mper;
      record.name = response[i].minm;
      record.itemCount = response[i].mimc;
      record.smartPlaylist = response[i].aeSP;

      result.appendElement(record, false);
    }

    return result;
  },
  
  /* void getItemsAsync (in sbIDaapItemsListener aListener, 
   *                      in long aDatabaseId, 
   *                      in nsIPropertyBag2 aProperties);
   */
  getItemsAsync: function (aListener, aDatabaseId, aProperties)
  {
    var params = this.getItemParameters(aProperties);
    var path = "/databases/"+ aDatabaseId + "/items?" + params;
    var handler = new AsyncItemsHandler();
    handler.init(aListener, this.uri, aDatabaseId);
    
    this.downloadAsync(path, handler);
  },

  /* void getPlaylistItemsAsync (in sbIDaapItemsListener aListener, 
   *                              in long aDatabaseId, in long aPlaylistId,
   *                              in nsIPropertyBag2 aProperties);
   */
  getPlaylistItemsAsync: function (aListener, aDatabaseId, 
                                    aPlaylistId, aProperties)
  {
    var params = this.getItemParameters(aProperties);
    var path = "/databases/"+ aDatabaseId + 
                "/containers/" + aPlaylistId + "/items?" + params;
    var handler = new AsyncItemsHandler();
    handler.init(aListener, this.uri, aDatabaseId, aPlaylistId);
    
    this.downloadAsync(path, handler);
  },

  /* void batchCreateMediaItems (in PRUint32 aDatabaseId, 
   *                             in nsIPropertyBag2 aProperties, 
   *                             in sbILibrary aLibrary, 
   *                             in sbIBatchCreateMediaItemsListener aListener,
   *                             [optional] in PRUint32 aPace);
   **/
  batchCreateMediaItems: function (aDatabaseId, aProperties, aLibrary, 
                                    aListener, aPace)
  {
    var listener = new BatchCreateListener(aLibrary, aListener, aPace);
    this.getItemsAsync(listener, aDatabaseId, aProperties);
  },

  /* private --------------------------------------------------------------*/

  getItemParameters : function(aProperties){
    var params = "type=music&meta=dmap.itemid,daap.songformat";

    if (aProperties) {
      if (aProperties.get(SBProperties.trackName))  params += ",dmap.itemname";
      if (aProperties.get(SBProperties.artistName)) params += ",daap.songartist";
      if (aProperties.get(SBProperties.albumName))  params += ",daap.songalbum";
      if (aProperties.get(SBProperties.albumArtistName))  params += ",daap.songalbumartist";
      if (aProperties.get(SBProperties.duration))   params += ",daap.songtime"; 
      if (aProperties.get(SBProperties.contentLength))  params += ",daap.songsize";
      if (aProperties.get(SBProperties.genre))    params += ",daap.songgenre";
      if (aProperties.get(SBProperties.bpm))      params += ",daap.songbeatsperminute";
      if (aProperties.get(SBProperties.bitRate))  params += ",daap.songbitrate";
      if (aProperties.get(SBProperties.comment))  params += ",daap.songcomment";
      if (aProperties.get(SBProperties.isPartOfCompilation)) params += ",daap.songcompilation";
      if (aProperties.get(SBProperties.created)) params += ",daap.songdateadded";
      if (aProperties.get(SBProperties.updated)) params += ",daap.songdatemodified";
      if (aProperties.get(SBProperties.totalDiscs)) params += ",daap.songdisccount";
      if (aProperties.get(SBProperties.discNumber)) params += ",dmap.songdiscnumber";
  //    if (aProperties.get(SBProperties.)) params += ",daap.songdisabled";
  //    if (aProperties.get(SBProperties.)) params += ",daap.songeqpreset";
  //    if (aProperties.get(SBProperties.)) params += ",daap.songdescription";
  //    if (aProperties.get(SBProperties.)) params += ",daap.songrelativevolume";
      if (aProperties.get(SBProperties.sampleRate)) params += ",daap.songsamplerate";
  //    if (aProperties.get(SBProperties.)) params += ",daap.songstarttime";
  //    if (aProperties.get(SBProperties.)) params += ",daap.songstoptime";
      if (aProperties.get(SBProperties.totalTracks)) params += ",daap.songtrackcount";
      if (aProperties.get(SBProperties.trackNumber)) params += ",daap.songtracknumber";
      if (aProperties.get(SBProperties.rating)) params += ",daap.songuserrating";
      if (aProperties.get(SBProperties.year))   params += ",daap.songyear";
  //    if (aProperties.get(SBProperties.)) params += ",daap.songdatakind";
      if (aProperties.get(SBProperties.originURL)) params += ",daap.songurl";
      if (aProperties.get(SBProperties.composerName)) params += ",daap.songcomposer";
    }

    return params + ",com.apple.itunes.norm-volume"
  },

  download: function (aPath, aContentHandler)
  {
    this.channel = null;

    if (!this.uri) {
      throw Cr.NS_ERROR_NOT_INITIALIZED;
    }

    var url = this.uri.resolve(aPath);
    var ioService = Cc["@mozilla.org/network/io-service;1"]
                      .getService(Ci.nsIIOService);

    var channel = ioService.newChannel(url, null, null);
    var stream = channel.open();

    /* save the channel */
    this.channel = channel.QueryInterface(Ci.nsIHttpChannel);

    if (!aContentHandler) {
      /* no handler => no parse =>  we're done */
      return null;
    }

    if (!this.channel || this.channel.responseStatus != 200) {
      /* no success, no parse */
      return null;
    }

    /* Create the Daap Parser */
    var parser = Cc["@songbirdnest.com/daap/parser;2"]
                  .createInstance(Ci.sbIDaapParser);

    parser.parseFromStream(stream, aContentHandler);
    stream.close();

    return aContentHandler.result;
  },

  //tmp, public: void downloadAsync (in AString aPath, in sbIDaapContentHandler aHandler); 
  downloadAsync: function (aPath, aContentHandler)
  {
    this.channel = null;

    if (!this.uri) {
      throw Cr.NS_ERROR_NOT_INITIALIZED;
    }

    var url = this.uri.resolve(aPath);
    var ioService = Cc["@mozilla.org/network/io-service;1"]
                      .getService(Ci.nsIIOService);

    var channel = ioService.newChannel(url, null, null);

    /* Create the Daap Parser */
    var parser = Cc["@songbirdnest.com/daap/parser;2"]
                  .createInstance(Ci.sbIDaapParser);
    parser.parseAsync(aContentHandler);

    channel.asyncOpen(parser, null);

    /* save the channel */
    this.channel = channel.QueryInterface(Ci.nsIHttpChannel);
  }

};



/*=========================================================================*/
/* Daap Server Info Response                                               */
/*=========================================================================*/

/* constructor */
function sbDaapServerInfo() {
}

/* implementation */
sbDaapServerInfo.prototype = {
  /* nsISupports ---------------------------------------------------------*/
  QueryInterface: XPCOMUtils.generateQI([Ci.sbIDaapServerInfo, 
                                          Ci.nsIClassInfo]),

  
  /* nsIClassInfo, so callers don't have to QI from JS -------------------*/
  classDescription: "Daap Server Info",
  classID:          Components.ID("{0330c3a2-bf3e-4074-a640-ab6e8a92ba5a}"),
  contractID:       "@songbirdnest.com/daap/server-info;1",
  flags                   : Ci.nsIClassInfo.THREADSAFE,
  implementationLanguage  : Ci.nsIProgrammingLanguage.JAVASCRIPT,
  getHelperForLanguage    : function(aLanguage) { return null; },
  getInterfaces : function(count) {
    var interfaces = [Ci.sbIDaapServerInfo, Ci.nsISupports, Ci.nsIClassInfo];
    count.value = interfaces.length;
    return interfaces;
  },


  /* sbIDaapServerInfo ----------------------------------------------------*/
  infoStatus: null,             //int
  protocol: null,               //version
  supportIndexing: null,        //byte
  supportExtensions: null,      //byte
  supportUpdate: null,          //byte
  supportAutoLogout: null,      //byte
  timeoutInterval: null,        //int
  loginRequired: null,          //byte
  supportQueries: null,         //byte
  serverName: null,             //string
  supportResolve: null,         //byte
  supportBrowsing: null,        //byte
  supportPersistentID: null,    //byte
  dmapVersion: null,            //verion

};

/*=========================================================================*/
/* Daap Content Code                                                       */
/*=========================================================================*/

/* constructor */
function sbDaapContentCode() {
}

/* implementation */
sbDaapContentCode.prototype = {
  /* nsISupports ----------------------------------------------------------*/
  QueryInterface: XPCOMUtils.generateQI([Ci.sbIDaapContentCode, 
                                          Ci.nsIClassInfo]),

  
  /* nsIClassInfo, so callers don't have to QI from JS --------------------*/
  classDescription: "Daap Content Code",
  classID:          Components.ID("{7066f287-15a0-4277-b590-f7992083307a}"),
  contractID:       "@songbirdnest.com/daap/content-code;1",
  flags                   : Ci.nsIClassInfo.THREADSAFE,
  implementationLanguage  : Ci.nsIProgrammingLanguage.JAVASCRIPT,
  getHelperForLanguage    : function(aLanguage) { return null; },
  getInterfaces : function(count) {
    var interfaces = [Ci.sbIDaapContentCode, Ci.nsISupports, Ci.nsIClassInfo];
    count.value = interfaces.length;
    return interfaces;
  },


  /* bIDaapContentCode ----------------------------------------------------*/
  code: null,           // mcnm
  name: null,           // minm
  type: null,           // mcty
};


/*=========================================================================*/
/* Daap Database                                                           */
/*=========================================================================*/

/* constructor */
function sbDaapDatabase() {
}

/* implementation */
sbDaapDatabase.prototype = {
  /* nsISupports ----------------------------------------------------------*/
  QueryInterface: XPCOMUtils.generateQI([Ci.sbIDaapDatabase, Ci.nsIClassInfo]),

  
  /* nsIClassInfo, so callers don't have to QI from JS --------------------*/
  classDescription: "Daap Database Response",
  classID:          Components.ID("{a58535f3-ce6f-4cd2-ae36-52fb31081587}"),
  contractID:       "@songbirdnest.com/daap/database-response;1",
  flags                   : Ci.nsIClassInfo.THREADSAFE,
  implementationLanguage  : Ci.nsIProgrammingLanguage.JAVASCRIPT,
  getHelperForLanguage    : function(aLanguage) { return null; },
  getInterfaces : function(count) {
    var interfaces = [Ci.sbIDaapDatabase, Ci.nsISupports, Ci.nsIClassInfo];
    count.value = interfaces.length;
    return interfaces;
  },


  /* sbIDaapDatabase ------------------------------------------------------*/
  id: null,
  persistentId: null,
  name: null,
  itemCount: null,
  containerCount: null
};


/*=========================================================================*/
/* Daap Playlist                                                           */
/*=========================================================================*/

/* constructor */
function sbDaapPlaylist() {
}

/* implementation */
sbDaapPlaylist.prototype = {
  /* nsISupports ----------------------------------------------------------*/
  QueryInterface: XPCOMUtils.generateQI([Ci.sbIDaapPlaylist, 
                                          Ci.nsIClassInfo]),

  
  /* nsIClassInfo, so callers don't have to QI from JS --------------------*/
  classDescription: "Daap Playlist Response",
  classID:          Components.ID("{d7cee193-4fc4-4a73-861e-793478c0bdcd}"),
  contractID:       "@songbirdnest.com/daap/playlist-response;1",
  flags                   : Ci.nsIClassInfo.THREADSAFE,
  implementationLanguage  : Ci.nsIProgrammingLanguage.JAVASCRIPT,
  getHelperForLanguage    : function(aLanguage) { return null; },
  getInterfaces : function(count) {
    var interfaces = [Ci.sbIDaapPlaylist, Ci.nsISupports, Ci.nsIClassInfo];
    count.value = interfaces.length;
    return interfaces;
  },


  /* sbIDaapPlaylist ------------------------------------------------------*/
  id: null,             // miid
  persistentId: null,   // mper
  name: null,           // minm
  itemCount: null,      // mimc
  smartPlaylist: null,  // aeSP
};


/*=========================================================================*/
/* Container Generic Content Handler                                       */
/*=========================================================================*/

function ContainerHandler ()
{
  this.result = null;
  this.current = null;
  this.text = null;

  this.tag = "mlit"; //default
}

ContainerHandler.prototype = { 
  result: null,  
  current: null,
  text: null,
  tag: null, //container searched

  startDocument: function() 
  {
    this.result = [];  
    this.current = null;
    this.text = null;
  },

  endDocument: function() {
    this.current = null;
    this.text = null;
  },

  startElement: function(aCode, aLength, aType) 
  {
    this.text = "";
    if (aCode == this.tag) {
      this.current = {};
    }
  },

  endElement: function(aCode, aLength, aType) 
  {
    if (aCode == this.tag) {
      this.result.push(this.current);
      this.current = null;
    }
    else if (this.current != null) {
      this.current[aCode] = this.text.replace(/^\s+|\s+$/g, "");
    }

    this.text = null;
  },

  characters: function(value) 
  {
    this.text += value; 
  },

  // nsISupports
  QueryInterface: XPCOMUtils.generateQI([Ci.sbIDaapContentHandler])

};


/*=========================================================================*/
/* Simple Attributes Generic Content Handler                               */
/*=========================================================================*/

function SimpleHandler ()
{
  this.result = null;
  this.text = null;
}

SimpleHandler.prototype = { 
  result: null,  
  text: null,

  startDocument: function() 
  {
    this.result = [];  
    this.text = null;
  },

  endDocument: function() {
    this.text = null;
  },

  startElement: function(aCode, aLength, aType) 
  {
    this.text = "";
  },

  endElement: function(aCode, aLength, aType) 
  {
    if (this.text) {
      this.result[aCode] = this.text.replace(/^\s+|\s+$/g, "");
    }
    this.text = null;
  },

  characters: function(value) 
  {
    this.text += value; 
  },

  // nsISupports
  QueryInterface: XPCOMUtils.generateQI([Ci.sbIDaapContentHandler])

};


/*=========================================================================*/
/* Asynchronic Daap Items Content Handler                                  */
/*=========================================================================*/

function AsyncItemsHandler ()
{
  this.listener = null;
  this.itemsBaseURL = null;

  this.item = null;
  this.properties = null;
  this.count = null;

  this.text = null;
}

AsyncItemsHandler.prototype = { 

  /* private data ---------------------------------------------------------*/
  listener: null,  
  itemsBaseURL: null,

  item: null,
  properties: null,
  count: null,

  text: null,

  /* public member --------------------------------------------------------*/
  init: function (aListener, aURI, aDatabaseId)
  {
    this.listener = aListener;
    this.itemsBaseURL = aURI.resolve("/databases/" + aDatabaseId + "/items");
  },

  /* sbIDaapContentHandler ------------------------------------------------*/
  startDocument: function() 
  {
    this.count = 0;
    this.listener.onEnumerationBegin();
  },

  endDocument: function() 
  {
    this.listener.onEnumerationEnd(this.count);

    /* clean */
    this.listener = null;
    this.itemsBaseURL = null;

    this.item = null;
    this.properties = null;
    this.count = null;

    this.text = null;
  },

  startElement: function(aCode, aLength, aType) 
  {
    this.text = "";
    if (aCode == "mlit") {
      this.item = {};
      this.properties = 
        Cc["@songbirdnest.com/Songbird/Properties/MutablePropertyArray;1"]
          .createInstance(Ci.sbIMutablePropertyArray);
    }
  },

  endElement: function(aCode, aLength, aType) 
  {
    if (aCode == "mlit") {
      if (this.item.id && this.item.format) {
        this.count++;
        var itemURL = this.itemsBaseURL + "/" 
                      + this.item.id + "." + this.item.format;

        try {
          this.listener.onEnumeratedItem(itemURL, this.properties)
        }
        catch (e if e.result == Cr.NS_ERROR_ABORT) {
          throw Cr.NS_ERROR_ABORT;
        }
      }
      /* FIXME: temporary fix because rhythmbox doesn't return song format 
       * for container items */
      else if (this.item.id) {
        /* send only the properties */
        try {
          this.listener.onEnumeratedItem(null, this.properties)
        }
        catch (e if e.result == Cr.NS_ERROR_ABORT) {
          throw Cr.NS_ERROR_ABORT;
        }
      }
      this.item = null;
      this.properties = null;
    }
    else if (this.item && this.text) {
      var value = this.text.replace(/^\s+|\s+$/g, "");
    
      switch (aCode) {
        case "miid": //dmap.itemid
          this.item.id = value; 
          //DProperty.id
          this.properties.appendProperty(
            "http://songbirdnest.com/data/1.0#daapId", value);
          break;
        case "asfm": this.item.format = value; break; //daap.songformat

        case "minm": //dmap.itemname
          this.properties.appendProperty(SBProperties.trackName, value);
          break;
        case "asar": //daap.songartist
          this.properties.appendProperty(SBProperties.artistName, value);
          break;
        case "asal": //daap.songalbum
          this.properties.appendProperty(SBProperties.albumName, value);
          break;
        case "asaa": //daap.songalbumartist
          this.properties.appendProperty(SBProperties.albumArtistName, value);
          break;
        case "astm": //daap.songtime
          if (value > 0) 
            this.properties.appendProperty(SBProperties.duration, (1000*value));
          break;
        case "assz": //daap.songsize
          if (value > 0) 
            this.properties.appendProperty(SBProperties.contentLength, value);
          break;
        case "asgn": //daap.songgenre
          this.properties.appendProperty(SBProperties.genre, value);
          break;
        case "asbt": //daap.songbeatsperminute
          if (value >= 0 && value <= 999) {
            this.properties.appendProperty(SBProperties.bpm, value);
          }
          break;
        case "asbr": //daap.songbitrate
          if (value > 0) 
            this.properties.appendProperty(SBProperties.bitRate, value);
          break;
        case "ascm": //daap.songcomment
          this.properties.appendProperty(SBProperties.comment, value);
          break;
        case "asco": //daap.songcompilation
          if (value > 0) 
            this.properties.appendProperty(SBProperties.isPartOfCompilation, 1);
          else  
            this.properties.appendProperty(SBProperties.isPartOfCompilation, 0);
          break;
        case "asda": //daap.songdateadded
          //this.properties.appendProperty(SBProperties.created, value);
          break;
        case "asdm": //daap.songdatemodified
          //this.properties.appendProperty(SBProperties.updated, value);
          break;
        case "asdc": //daap.songdisccount
          if (value > 0 && value <= 999) 
            this.properties.appendProperty(SBProperties.totalDiscs, value);
          break;
        case "asdn": //dmap.songdiscnumber
          if (value > 0 && value <= 999) 
            this.properties.appendProperty(SBProperties.discNumber, value);
          break;
        case "asdb": //daap.songdisabled
          //this.properties.appendProperty(, value);
          break;
        case "aseq": //daap.songeqpreset
          //this.properties.appendProperty(, value);
          break;
        case "asdt": //daap.songdescription
          //this.properties.appendProperty(, value);
          break;
        case "asrv": //daap.songrelativevolume
          //this.properties.appendProperty(, value);
          break;
        case "assr": //daap.songsamplerate
          if (value > 0) 
            this.properties.appendProperty(SBProperties.sampleRate, value);
          break;
        case "asst": //daap.songstarttime
          //this.properties.appendProperty(, value);
          break;
        case "assp": //daap.songstoptime
          //this.properties.appendProperty(, value);
          break;
        case "astc": //daap.songtrackcount
          if (value > 0 && value <= 999) 
            this.properties.appendProperty(SBProperties.totalTracks, value);
          break;
        case "astn": //daap.songtracknumber
          if (value > 0 && value <= 999) 
            this.properties.appendProperty(SBProperties.trackNumber, value);
          break;
        case "asur": //daap.songuserrating
          if (value >= 0 && value <= 100) 
            this.properties.appendProperty(SBProperties.rating, Math.round(value/20));
          break;
        case "asyr": //daap.songyear
          if (value >= 0 && value <= 9999) 
            this.properties.appendProperty(SBProperties.year, value);
          break;
        case "asdk": //daap.songdatakind
          //this.properties.appendProperty(, value);
          break;
        case "asul": //daap.songurl
/*
          try { 
            this.properties.appendProperty(SBProperties.originURL, value); 
          } catch (e) {}
          break;
*/
        case "ascp": //daap.songcomposer
          this.properties.appendProperty(SBProperties.composerName, value);
          break;
      }
    }

    this.text = null;
  },

  characters: function(value) 
  {
    this.text += value; 
  },

  /* nsISupports ----------------------------------------------------------*/
  QueryInterface: XPCOMUtils.generateQI([Ci.sbIDaapContentHandler])

};


/*=========================================================================*/
/* Batch Create Media Items Helper Listener                                */
/*=========================================================================*/

function BatchCreateListener(aLibrary, aListener, aPace)
{
  if (!(aLibrary instanceof Ci.sbILibrary)) {
    throw Cr.NS_ERROR_INVALID_ARG;
  }

  this.library = aLibrary;

  if (aListener instanceof Ci.sbIBatchCreateMediaItemsListener) {
    this.listener = aListener;
  }
  else {
    this.listener = null;
  }

  if (aPace > 0) {
    this.pace = aPace;
  }
  else {
    this.pace = 100;
  }

  this.count = null;
  this.mediaItems = null;
  this.trackArray = null;
  this.propertiesArray = null;
  this.ioService = null;
  this.stop = null;
  this.abort = null;
}

BatchCreateListener.prototype = {

  /* private data ---------------------------------------------------*/
  library: null,
  listener: null,
  pace: null,
  count: null,
  mediaItems: null,
  trackArray: null,
  propertiesArray: null,
  ioService: null,
  stop: null,
  abort: null,

  /* nsISupports ----------------------------------------------------*/
  QueryInterface: XPCOMUtils.generateQI([Ci.sbIDaapItemsListener,
                                Ci.sbIBatchCreateMediaItemsListener]),

  /* sbIDaapItemsListener -------------------------------------------*/
  onEnumerationBegin: function() 
  {
    this.count = 0;
    this.stop = false;
    this.abort = false;
    this.start = Date.now();

    /* get the IO service */
    this.ioService = Cc["@mozilla.org/network/io-service;1"]
                        .getService(Ci.nsIIOService);

    if (!this.ioService) {
      throw Components.results.NS_ERROR_FAILURE;
    }

    this.mediaItems = Cc["@songbirdnest.com/moz/xpcom/threadsafe-array;1"]
                        .createInstance(Ci.nsIMutableArray);
    
    this.createArrays();
  },

  onEnumeratedItem: function (aItemURL, aProperties) 
  {
    if (this.abort) {
      //abort
      //throw Cr.NS_ERROR_ABORT;  <---------- this doesn't work XXX
      throw Components.Exception(
        "'Abort' when calling method: [sbIDaapContentHandler::endElement]", 
        Cr.NS_ERROR_ABORT);
    }

    var index = this.trackArray.length;
    var uri = this.ioService.newURI(aItemURL, null, null);

    this.trackArray.insertElementAt(uri, index, false);
    this.propertiesArray.insertElementAt(aProperties, index, false);
    index++;


    if (index == this.pace) {
      this.createItems();

      /*  be a little "polite" an let others work */
      /*
       * this is in the Job now.
       *
      let thread = Cc["@mozilla.org/thread-manager;1"]
                      .getService().currentThread;
      for (let i=0; i<50 && thread.hasPendingEvents(); i++) {
        thread.processNextEvent(false);
      }
      */
    }
  },

  onEnumerationEnd: function(aCount)
  {
    this.stop = true;

    if (this.trackArray.length) {
      this.createItems();
    }
  },

  /* sbIBatchCreateMediaItemsListener -------------------------------------*/
  onProgress: function(aIndex)
  {
    /* not used */
  },

  onComplete: function(aMediaItems, aResult)
  {
    if (!this.count 
        || !this.mediaItems
        || this.count <= this.mediaItems.length
        || this.abort) {
      //error, finished or aborted: do nothing
      return;
    }

    /* mediaItems created */
    if (aResult == Cr.NS_OK) {
      for (var i=0; i<aMediaItems.length; i++) {
        var item = aMediaItems.queryElementAt(i, Ci.nsISupports);
        this.mediaItems.appendElement(item, false);
      }

      if (this.listener) {
        //XXX: while calling this listener may occur
        // a context switch to other threads !!
        try { 
          this.listener.onProgress(this.mediaItems.length);
        }
        catch (e if e.result == Cr.NS_ERROR_ABORT) {
          this.abort = true;
        }
        if (this.abort) {
          //notify the listener that we just aborted
          try {
            this.listener.onComplete(null, Cr.NS_ERROR_ABORT);
          }
          catch (e) {}
        }
      }
    }
    else {
      /* error */
      if (this.listener) {
        this.listener.onComplete(null, aResult);
      }
      return ;
    }

    /* re-check the stop condition */ 
    // XXX: don't remnove this!!! 
    //   (note the last condition is < not <= as in the top) 
    if (!this.count 
        || !this.mediaItems
        || this.count < this.mediaItems.length) {
      //error or finished: do nothing
      return;
    }

    /* finished */
    if (this.stop && this.count == this.mediaItems.length) {
      var mediaItems = this.mediaItems;

      /* clean */
      this.library = null;
      this.pace = null;
      this.count = null;
      this.mediaItems = null;
      this.trackArray = null;
      this.propertiesArray = null;
      this.ioService = null;

      // let do this a the end to avoid context switch
      //  between the call and the clean up
      if (this.listener) {
        this.listener.onComplete(mediaItems, Cr.NS_OK);
      }
    }
  },

  /* private functions ----------------------------------------------------*/
  createItems: function ()
  {
    this.count += this.trackArray.length;

    this.library.batchCreateMediaItemsAsync(
              this, 
              this.trackArray, 
              this.propertiesArray, 
              false);

    /* new arrays */        
    this.createArrays();
  },

  createArrays: function ()
  {
    /* create the track and the properties arrays needed for the async call */
    this.trackArray = Cc["@songbirdnest.com/moz/xpcom/threadsafe-array;1"]
                        .createInstance(Ci.nsIMutableArray);
    this.propertiesArray = Cc["@songbirdnest.com/moz/xpcom/threadsafe-array;1"]
                              .createInstance(Ci.nsIMutableArray);

    if (!this.trackArray || !this.propertiesArray) {
      this.trackArray = null;
      this.propertiesArray = null;

      throw Components.results.NS_ERROR_FAILURE;
    }
  }

};


/*=========================================================================*/
/* XPCOM Module                                                            */
/*=========================================================================*/

/* components constructors */
var components = [sbDaapClient, sbDaapServerInfo, 
                    sbDaapDatabase, sbDaapPlaylist];

/* NSGetModule entry point */
function NSGetModule(compMgr, fileSpec) {
  return XPCOMUtils.generateModule(components);
}

