/*
//
// BEGIN SONGBIRD GPL
//
// This file is part of the Songbird web player.
//
// Copyright(c) 2005-2008 POTI, Inc.
// http://songbirdnest.com
//
// This file may be licensed under the terms of of the
// GNU General Public License Version 2 (the "GPL").
//
// Software distributed under the License is distributed
// on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, either
// express or implied. See the GPL for the specific language
// governing rights and limitations.
//
// You should have received a copy of the GPL along with this
// program. If not, go to http://www.gnu.org/licenses/gpl.html
// or write to the Free Software Foundation, Inc.,
// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
//
// END SONGBIRD GPL
//
*/

const Ci = Components.interfaces;
const Cc = Components.classes;
const Cr = Components.results;
const Cu = Components.utils;

const nsISupports = Components.interfaces.nsISupports;

const CLASS_ID = Components.ID("4114d18b-497f-48c6-87f8-ba47877cc7ff");
const CLASS_NAME = "Daap Session Manager";
const CONTRACT_ID = "@songbirdnest.com/daap/session-manager;1";


Components.utils.import("resource://gre/modules/XPCOMUtils.jsm");

/* SbDaapSessionManager constructor ---------------------------------------*/
function SbDaapSessionManager() 
{

}

/* SbDaapSessionManager implementation ------------------------------------*/
SbDaapSessionManager.prototype = {

  /* nsISupports ----------------------------------------------------------*/
  QueryInterface: function(aIID)
  {
    if (!aIID.equals(Ci.sbIDaapSessionManager)
        && !aIID.equals(nsISupports))
        throw Components.results.NS_ERROR_NO_INTERFACE;
    return this;
  },

  /* private data ---------------------------------------------------------*/
  sessions: {},
  dm: null,
  seed_v2: null,
  seed_v3: null,

  /* sbIDaapSessionManager methods ----------------------------------------*/

  /* PRUint32 initSession(in sbIDaapClient aClient, 
   *                      [optional] in AString aPassword); 
   **/
  initSession: function(aClient, aPassword)
  {
    if (!aClient) {
      throw Cr.NS_ERROR_INVALID_ARG;
    }

    var uri = aClient.URI;
    var key = this.getKey(uri);
    
    /*
    if (!aClient) {
      throw Cr.NS_ERROR_FAILURE;
    }
    aClient.init(uri);
    */

    if (this.sessions[key] && this.sessions[key].id) {
      /* already opened session, make a simple query for validation */
      //XXXDaniel: don't use /update because some server may not support it 
      //           e.g. rhythmbox
      //XXXDaniel: don't use /server-info or /content-codes because login
      //           isn't required
      aClient.getDatabases();

      if (aClient.lastHttpChannel.responseStatus == 200) {
        return this.sessions[key].id;
      }

      /* else invalidate the session */
      this.sessions[key] = undefined;
    }

    /* make a new session */
    this.sessions[key] = {}

    /* remember the password */
    if (uri.password.length > 0) {
      this.sessions[key].password = uri.password;
    }
    if (aPassword && aPassword.length > 0) {
      this.sessions[key].password = aPassword;
    }

    /* make sure we have this server codes */
    /* FIXME: deprecated??
    var codesManager = Cc["@songbirdnest.com/daap/codes-manager;1"]
                        .getService(Ci.sbIDaapCodesManager);
    if (codesManager) {
      codesManager.downloadCodes(aClient);
    }

    /* get the server version to use the rigth hash */
    var serverInfo = aClient.getServerInfo();
    if (serverInfo) {
      this.sessions[key].version = parseFloat(serverInfo.protocol);
    }
    
    var sessionId = aClient.login(null, true);
    this.sessions[key].id = sessionId;

    if (aClient.lastHttpChannel.responseStatus != 200) {
      /* abort, let other handle it */
      this.sessions[key] = undefined;

      /* XXX review this */
      if (aClient.lastHttpChannel.responseStatus == 401) {
        /* auth fail, ask the user for a password */
        let auth = Cc["@songbirdnest.com/daap/auth-prompt;1"]
                    .createInstance(Ci.nsIAuthPrompt2);
        if (auth) {
          var authInfo = {
              QueryInterface: XPCOMUtils.generateQI([Ci.nsIAuthInformation]),
              password: null
            };

          if (auth.promptAuth(aClient.lastHttpChannel, 
                              Ci.nsIAuthPrompt2.LEVEL_PW_ENCRYPTED, 
                              authInfo)) {
            return this.initSession(aClient, authInfo.password);
          }
        }
      }
      else if (this.handleStatus(aClient.lastHttpChannel.responseStatus)) {
        /* retry */
        return this.initSession(aClient, aPassword);
      }

      throw Cr.NS_ERROR_FAILURE;
    }

    /* initialize the request id */ 
    this.sessions[key].requestId = 0;

    return this.sessions[key].id;
  },

  /*  PRUint32 getSession(in nsIURI aURI); */
  getSession: function(aURI)
  {
    var key = this.getKey(aURI);
    if (this.sessions[key] && this.sessions[key].id) {
      return this.sessions[key].id;
    }

    /* no session found */
    throw Cr.NS_ERROR_NOT_AVAILABLE;
  },

  /* void closeSession(in sbIDaapClient aClient); */
  closeSession: function(aClient)
  {
    var uri = aClient.URI;
    var key = this.getKey(uri);

    try {
      /* try to close the session first */
      aClient.init(uri);
      aClient.logout(true);
    }
    catch (e) {}

    if (this.sessions[key]) {
      this.sessions[key] = undefined;
    }
  },

  closeAll: function()
  {
    var client = Cc["@songbirdnest.com/daap/client;1"]
                   .createInstance(Ci.sbIDaapClient);
   var ioService = Cc["@mozilla.org/network/io-service;1"]
                    .getService(Ci.nsIIOService);

    for (var key in this.sessions) {
      if (this.sessions[key].id) {
        var uri = ioService.newURI(key, null, null);
        client.init(uri);
        this.closeSession(client);
      }
    }

  },

  /*  nsIHttpChannel addRequestHeaders(in nsIURI aURI, in nsIHttpChannel aChannel); */
  addRequestHeaders: function(aURI, aChannel)
  {
    var key = this.getKey(aURI);
    if (!this.sessions[key]) {
      /* no session, no headers */
      return aChannel;
    }

    /* auth headers */
    if (this.sessions[key].password) {
      var b64 = btoa("user:" + this.sessions[key].password)
      aChannel.setRequestHeader("Authorization", "Basic "+b64, false);
    }
    
    if (this.sessions[key].id) {
      this.sessions[key].requestId += 1;

      aChannel.setRequestHeader("Client-DAAP-Request-ID", 
                                this.sessions[key].requestId, false);

      var hash = this.getHash(aURI, this.sessions[key].requestId);
      aChannel.setRequestHeader("Client-DAAP-Validation", hash, false);
    }

    return aChannel;
  },

  /* private methods ------------------------------------------------------*/

  getKey: function(aURI)
  {
    return (aURI.scheme + "://" + aURI.host);
  },

  getHash: function(aURI, aSequence)
  {
    var key = this.getKey(aURI);
    var select = 2;
    var seed;
    var crypto; 

    if (this.sessions[key].version == 2) {
      seed = this.getSeedv2();
      crypto = Cc["@mozilla.org/security/hash;1"]
                  .createInstance(Ci.nsICryptoHash);
      crypto.init(crypto.MD5);
    }
    else if (this.sessions[key].version > 2) {
      seed = this.getSeedv3();
      crypto = Cc["@songbirdnest.com/daap/cryptohash;1"]
                  .createInstance(Ci.sbIDaapCryptoHash);
      crypto.init(crypto.MD5DAAP);
    }

    var result;

    result = this.toArray(aURI.path);
    crypto.update(result.value, result.length);
    result = this.toArray("Copyright 2003 Apple Computer, Inc.");
    crypto.update(result.value, result.length);
    result = this.toArray(seed[select]);
    crypto.update(result.value, result.length);

    if (this.sessions[key].version > 2 && aSequence > 0) {
      result = this.toArray(""+aSequence);
      crypto.update(result.value, result.length);
    }
   
    // convert the binary hash data to a hex string.
    var hash = crypto.finish(false);
    var s = [this.toHexString(hash.charCodeAt(i)) for (i in hash)].join("");
 
    return s.toUpperCase();
  },

  getSeedv2: function()
  {
    /* MD5 */

    if (this.seed_v2) {
      return this.seed_v2;
    }

    this.seed_v2 = [];

    //for (var i=0; i<255; i++) {
    for (var i=0; i<3; i++) {
      var crypto = Cc["@mozilla.org/security/hash;1"]
                  .createInstance(Ci.nsICryptoHash);
      crypto.init(crypto.MD5);

      var text;
      var result;
      
      text = (i & 0x80) ? "Accept-Language" : "user-agent";
      result = this.toArray(text);
      crypto.update(result.value, result.length);

      text = (i & 0x40) ? "max-age" : "Authorization";
      result = this.toArray(text);
      crypto.update(result.value, result.length);
      
      text = (i & 0x20) ? "Client-DAAP-Version" : "Accept-Encoding";
      result = this.toArray(text);
      crypto.update(result.value, result.length);
      
      text = (i & 0x10) ? "daap.protocolversion" : "daap.songartist";
      result = this.toArray(text);
      crypto.update(result.value, result.length);
      
      text = (i & 0x08) ? "daap.songcomposer" : "daap.songdatemodified";
      result = this.toArray(text);
      crypto.update(result.value, result.length);
      
      text = (i & 0x04) ? "daap.songdiscnumber" : "daap.songdisabled";
      result = this.toArray(text);
      crypto.update(result.value, result.length);
      
      text = (i & 0x02) ? "playlist-item-spec" : "revision-number";
      result = this.toArray(text);
      crypto.update(result.value, result.length);
      
      text = (i & 0x01) ? "session-id" : "content-codes";
      result = this.toArray(text);
      crypto.update(result.value, result.length);

      /* convert the binary hash data to a hex string. */
      var hash = crypto.finish(false);
      var s = [this.toHexString(hash.charCodeAt(i)) for (i in hash)].join("");
      
      this.seed_v2.push(s.toUpperCase());
    }
    //}

    return this.seed_v2;
  },

  getSeedv3: function()
  {
    /* apple MD5 */

    if (this.seed_v3) {
      return this.seed_v3;
    }

    this.seed_v3 = [];

    //for (var i=0; i<255; i++) {
    for (var i=0; i<3; i++) {
      var crypto = Cc["@songbirdnest.com/daap/cryptohash;1"]
                  .createInstance(Ci.sbIDaapCryptoHash);
      crypto.init(crypto.MD5DAAP);

      var text;
      var result;

      text = (i & 0x40) ? "eqwsdxcqwesdc" : "op[;lm,piojkmn";
      result = this.toArray(text);
      crypto.update(result.value, result.length);
      
      text = (i & 0x20) ? "876trfvb 34rtgbvc" : "=-0ol.,m3ewrdfv";
      result = this.toArray(text);
      crypto.update(result.value, result.length);
      
      text = (i & 0x10) ? "87654323e4rgbv " : "1535753690868867974342659792";
      result = this.toArray(text);
      crypto.update(result.value, result.length);
      
      text = (i & 0x08) ? "Song Name" : "DAAP-CLIENT-ID:";
      result = this.toArray(text);
      crypto.update(result.value, result.length);
      
      text = (i & 0x04) ? "111222333444555" : "4089961010";
      result = this.toArray(text);
      crypto.update(result.value, result.length);
      
      text = (i & 0x02) ? "playlist-item-spec" : "revision-number";
      result = this.toArray(text);
      crypto.update(result.value, result.length);
      
      text = (i & 0x01) ? "session-id" : "content-codes";
      result = this.toArray(text);
      crypto.update(result.value, result.length);
      
      text = (i & 0x80) ? "IUYHGFDCXWEDFGHN" : "iuytgfdxwerfghjm";
      result = this.toArray(text);
      crypto.update(result.value, result.length);

      /* convert the binary hash data to a hex string. */
      var hash = crypto.finish(false);
      var s = [this.toHexString(hash.charCodeAt(i)) for (i in hash)].join("");
      
      this.seed_v3.push(s.toUpperCase());
    }
    //}

    return this.seed_v3;
  },

  /* return the two-digit hexadecimal code for a byte */
  toHexString: function(charCode)
  {
    return ("0" + charCode.toString(16)).slice(-2);
  },

  /* convert a astring into an array */ 
  toArray: function (text)
  {
    var converter = Cc["@mozilla.org/intl/scriptableunicodeconverter"]
                      .createInstance(Ci.nsIScriptableUnicodeConverter);
    converter.charset = "UTF-8";

    var result = {value: null};
    var data = converter.convertToByteArray(text, result);

    result.length = result.value;
    result.value = data;

    return result;
  },

  handleStatus: function (aStatus)
  {
    /* status ok ------------------------*/
    if (aStatus >= 200 && aStatus < 300) {
      /* 204 means no content, e.g logout response */
      return true;
    }

    /* status error ---------------------*/
    if (aStatus == 400) {
      /* bad request, nothing to do */
      return false;
    }

    /* get the main window */
    if (typeof(window) == "undefined") { 
      var window = Cc["@mozilla.org/appshell/window-mediator;1"]
                    .getService(Ci.nsIWindowMediator)
                    .getMostRecentWindow("Songbird:Main").window;
    }

    /* prompt service */
    var prompts = Cc["@mozilla.org/embedcomp/prompt-service;1"]
                  .getService(Ci.nsIPromptService);
    let title = "Error";
    let message;

    if (aStatus == 401) {
      /* auth required */
      //this shouldn't be handled here
      return false;
    }

    // XXX TODO: localize these XXX
    if (aStatus == 403) {
      /* probably session expired, delete the current session */
      this.logout();
      message = "Invalid session. Retry?";
    }
    else if (aStatus == 503) {
      /* probably max connections to server */
      message = "Cannot connect to server, probably too much users connected. Retry?";
    }
    else if (aStatus == -1) {
      /* probably invalid URL */
      message = "Cannot connect to server. Retry?";
    }
    else {
      /* unknown error */
      message = "Unknown error. Retry?";
    }

    /* ask if retry */
    if (prompts.confirm(null, title, message)) {
      /* retry */
      return true;
    }

    return false;
  },  //handleStatus

}; //SbDaapSessionManager


//=================================================
// Note: You probably don't want to edit anything
// below this unless you know what you're doing.
//
// Factory
var SbDaapSessionManagerFactory = {
  singleton: null,
  createInstance: function (aOuter, aIID)
  {
    if (aOuter != null)
      throw Components.results.NS_ERROR_NO_AGGREGATION;
    if (this.singleton == null)
      this.singleton = new SbDaapSessionManager();
    return this.singleton.QueryInterface(aIID);
  }
};

// Module
var SbDaapSessionManagerModule = {
  registerSelf: function(aCompMgr, aFileSpec, aLocation, aType)
  {
    aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.registerFactoryLocation(CLASS_ID, CLASS_NAME, CONTRACT_ID, aFileSpec, aLocation, aType);
  },

  unregisterSelf: function(aCompMgr, aLocation, aType)
  {
    aCompMgr = aCompMgr.QueryInterface(Components.interfaces.nsIComponentRegistrar);
    aCompMgr.unregisterFactoryLocation(CLASS_ID, aLocation);        
  },
  
  getClassObject: function(aCompMgr, aCID, aIID)
  {
    if (!aIID.equals(Components.interfaces.nsIFactory))
      throw Components.results.NS_ERROR_NOT_IMPLEMENTED;

    if (aCID.equals(CLASS_ID))
      return SbDaapSessionManagerFactory;

    throw Components.results.NS_ERROR_NO_INTERFACE;
  },

  canUnload: function(aCompMgr) { return true; }
};

//module initialization
function NSGetModule(aCompMgr, aFileSpec) { return SbDaapSessionManagerModule; }

